#  Copyright 2025 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import torch
import torch.nn as nn
import timm
import safetensors.torch  
import os

class SmallDINOv2(nn.Module):
    def __init__(self, model_dir, use_safetensors=True):
        """
        Load DINOv2 ViT-S/14 model from local files using timm.

        Args:
            model_dir (str): Path to the directory containing the model weights.
            use_safetensors (bool): Whether to load from .safetensors (default) or .bin.
        """
        super(SmallDINOv2, self).__init__()

        # Load DINOv2 model architecture from timm
        self.model = timm.create_model(
            'vit_small_patch14_dinov2.lvd142m',
            pretrained=False,  # Do not download weights online
            num_classes=0  # Remove classifier
        )

        # Determine which weight file to use
        weights_path = os.path.join(model_dir, "model.safetensors" if use_safetensors else "pytorch_model.bin")
        if not os.path.exists(weights_path):
            raise FileNotFoundError(f"Model weights not found in {weights_path}")

        # Load weights
        if use_safetensors:
            state_dict = safetensors.torch.load_file(weights_path)
        else:
            state_dict = torch.load(weights_path, map_location="cpu")

        # Load weights into model (allow missing keys since classifier is removed)
        self.model.load_state_dict(state_dict, strict=False)
        print(f"Loaded model weights from {weights_path}")

        # Freeze all parameters (acts as a fixed feature extractor)
        for param in self.model.parameters():
            param.requires_grad = False  

    def forward(self, x):
        """
        Extracts patch-level features using DINOv2.
        
        Args:
            x (torch.Tensor): Input tensor of shape [B, 3, H, W] (single patch at a time).
        
        Returns:
            torch.Tensor: Extracted CLS token features [B, 384].
        """
        features = self.model.forward_features(x)  # Extract raw features
        
        # CLS token is usually at index 0 in ViT outputs
        return features[:, 0, :]  # Extract the CLS token representation



#  Copyright 2025 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
The following is the UNICORN baseline.

It is meant to run within a container.

To save the container and prep it for upload to Grand-Challenge.org you can call:

  ./do_save.sh

Any container that shows the same behaviour will do, this is purely an example of how one COULD do it.

Reference the documentation to get details on the runtime environment on the platform:
https://grand-challenge.org/documentation/runtime-environment/

Happy programming!
"""

import torch
import torchvision

torchvision.disable_beta_transforms_warning()

from pathlib import Path

from unicorn_baseline.io import load_inputs, load_task_description
from unicorn_baseline.language.main import run_language
from unicorn_baseline.vision.main import run_vision
from unicorn_baseline.vision_language.main import run_vision_language_task

INPUT_PATH = Path("/input")
OUTPUT_PATH = Path("/output")
MODEL_PATH = Path("/opt/ml/model")

def print_directory_contents(path: Path | str):
    path = Path(path)
    for child in path.iterdir():
        if child.is_dir():
            print_directory_contents(child)
        else:
            print(child)


def _show_torch_cuda_info():
    print("=+=" * 10)
    print("Collecting Torch CUDA information")
    print(f"Torch CUDA is available: {(available := torch.cuda.is_available())}")
    if available:
        print(f"- number of devices: {torch.cuda.device_count()}")
        print(f"- current device: { (current_device := torch.cuda.current_device())}")
        print(f"- properties: {torch.cuda.get_device_properties(current_device).name}")
    print("=+=" * 10)


def run_vision_and_visionlanguage(input_dir: Path, model_dir: Path) -> int:
    """
    Process input data
    """

    task_description = load_task_description(
        input_path=input_dir / "unicorn-task-description.json"
    )
    input_information = load_inputs(input_path=input_dir / "inputs.json")

    # retrieve task details
    domain = task_description["domain"]
    modality = task_description["modality"]
    task_type = task_description["task_type"]

    if modality == "vision":
        run_vision(
            task_description=task_description,
            input_information=input_information,
            model_dir=model_dir,
        )
    elif modality == "vision-language":
        run_vision_language_task(
            input_information=input_information,
            model_dir=model_dir,
        )
    else:
        raise ValueError(
            f"Modality '{modality}' and domain '{domain}' not supported yet"
        )

    return 0


def run():
    # show GPU information
    _show_torch_cuda_info()

    # print contents of input folder
    print("input folder contents:")
    print_directory_contents(INPUT_PATH)
    print("=+=" * 10)

    # check if the task is image or text
    if (INPUT_PATH / "nlp-task-configuration.json").exists():
        return run_language(OUTPUT_PATH)
    else:
        return run_vision_and_visionlanguage(INPUT_PATH, MODEL_PATH)


if __name__ == "__main__":

    raise SystemExit(run())



from lighter_zoo import SegResNet
from monai.transforms import (
    Compose,
    LoadImage,
    EnsureType,
    ScaleIntensityRange,
    Orientation,
    Resize,
)
from monai.data import Dataset, DataLoader
import torch.nn as nn
import numpy as np
import torch


def load_model(model_path):
    # Load pre-trained model
    return SegResNet.from_pretrained(model_path)


def load_data(data):
    train_transforms = Compose(
        [
            EnsureType(),
            ScaleIntensityRange(a_min=-1024, a_max=2048, b_min=0, b_max=1, clip=True),
        ]
    )

    train_ds = Dataset(data=data, transform=train_transforms)
    train_loader = DataLoader(train_ds, batch_size=1, shuffle=False)
    return train_loader


def encode(model, patches):
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    model.to(device)
    adaptive_pool = nn.AdaptiveAvgPool3d((1, 1, 1))

    # expand patch to match encoder input requirements
    patch_array = np.expand_dims(patches, axis=(0, 1))

    train_loader = load_data(patch_array)

    model.eval()
    with torch.no_grad():
        for input in train_loader:
            input = input.to(device)
            output = model.encoder(input)
            # average pool and flatten the output to fit feature vector requirements
            output_flat = adaptive_pool(output[-1])
            out_flat = output_flat.flatten(start_dim=0)

    return out_flat.cpu().detach().numpy().tolist()



def update_state_dict(
    *,
    model_dict: dict,
    state_dict: dict,
):
    """
    Matches weights between `model_dict` and `state_dict`, accounting for:
    - Key mismatches (missing in model_dict)
    - Shape mismatches (tensor size differences)

    Args:
        model_dict (dict): model state dictionary (expected keys and shapes)
        state_dict (dict): checkpoint state dictionary (loaded keys and values)

    Returns:
        updated_state_dict (dict): Weights mapped correctly to `model_dict`
        msg (str): Log message summarizing the result
    """
    success = 0
    shape_mismatch = 0
    missing_keys = 0
    updated_state_dict = {}
    shape_mismatch_list = []
    missing_keys_list = []
    used_keys = set()
    for model_key, model_val in model_dict.items():
        matched_key = False
        for state_key, state_val in state_dict.items():
            if state_key in used_keys:
                continue
            if model_key == state_key:
                if model_val.size() == state_val.size():
                    updated_state_dict[model_key] = state_val
                    used_keys.add(state_key)
                    success += 1
                    matched_key = True  # key is successfully matched
                    break
                else:
                    shape_mismatch += 1
                    shape_mismatch_list.append(model_key)
                    matched_key = True  # key is matched, but weight cannot be loaded
                    break
        if not matched_key:
            # key not found in state_dict
            updated_state_dict[model_key] = model_val  # Keep original weights
            missing_keys += 1
            missing_keys_list.append(model_key)
    # Log summary
    msg = f"{success}/{len(model_dict)} weight(s) loaded successfully"
    if shape_mismatch > 0:
        msg += f"\n{shape_mismatch} weight(s) not loaded due to mismatching shapes: {shape_mismatch_list}"
    if missing_keys > 0:
        msg += f"\n{missing_keys} key(s) from checkpoint not found in model: {missing_keys_list}"
    return updated_state_dict, msg



import os
from pathlib import Path

from unicorn_baseline.io import resolve_image_path, write_json_file
from unicorn_baseline.vision.pathology.main import extract_coordinates, save_coordinates
from unicorn_baseline.vision.pathology.wsi import FilterParams, TilingParams
from unicorn_baseline.vision_language.inference import generate_caption
from unicorn_baseline.vision_language.models import PRISM, Virchow
from transformers import MarianMTModel, MarianTokenizer


def get_file_path(file_location, extensions):
    for ext in extensions:
        potential_path = Path(f"{file_location}{ext}")
        if potential_path.exists():
            return potential_path
    return None


def save_output(caption, name):
    output_dict = [{"uid": name, "text": caption}]
    slug = "nlp-predictions-dataset"
    output_path = Path("/output")
    output_filename = output_path / f"{slug}.json"

    write_json_file(
        location=output_filename,
        content=output_dict,
    )

    print(f"Caption saved to {output_filename}")


def translate(caption, model_dir, model_name="opus-mt-en-nl"):
    """
    Translates text from English to Dutch using the Helsinki model. Note this model is just one example of a translation model that is publicly available.
    The model is available on HuggingFace https://huggingface.co/Helsinki-NLP/opus-mt-en-nl. Note that the model is not trained on medical data, so the translation may not be perfect. Alternative models may perform better.

    Args:
        caption (str): Caption to be translated.
        model_dir (str): Base directory where models are stored.
        model_name (str): Name of the translation model.

    Returns

    -------
        str: Caption translated to Dutch.
    """

    model_path = os.path.join(model_dir, model_name)

    # Assert the model path exists
    assert os.path.exists(model_path), f"Model path does not exist: {model_path}"

    tokenizer = MarianTokenizer.from_pretrained(model_path)
    model = MarianMTModel.from_pretrained(model_path)

    translated = model.generate(**tokenizer(caption, return_tensors="pt", padding=True))
    caption_translated = [
        tokenizer.decode(t, skip_special_tokens=True) for t in translated
    ][0]
    return caption_translated


def run_vision_language_task(*, input_information, model_dir):

    tissue_mask_path = None
    for input_socket in input_information:
        if input_socket["interface"]["kind"] == "Image":
            image_name = input_socket["image"]["name"]
            wsi_path = resolve_image_path(location=input_socket["input_location"])
        elif input_socket["interface"]["kind"] == "Segmentation":
            tissue_mask_path = resolve_image_path(
                location=input_socket["input_location"]
            )

    num_workers = 4
    batch_size = 32
    mixed_precision = True
    max_number_of_tiles = 14000
    tiling_params = TilingParams(
        spacing=0.5,
        tolerance=0.07,
        tile_size=224,
        overlap=0.0,
        drop_holes=False,
        min_tissue_ratio=0.25,
        use_padding=True,
    )
    filter_params = FilterParams(ref_tile_size=256, a_t=4, a_h=2, max_n_holes=8)

    # create output directories
    coordinates_dir = Path("/tmp/coordinates/")
    coordinates_dir.mkdir(parents=True, exist_ok=True)
    coordinates, _, level, resize_factor, tile_size_lv0, image_spacing, image_size = (
        extract_coordinates(
            wsi_path=wsi_path,
            tissue_mask_path=tissue_mask_path,
            tiling_params=tiling_params,
            filter_params=filter_params,
            max_number_of_tiles=max_number_of_tiles,
            num_workers=num_workers,
        )
    )

    save_coordinates(
        wsi_path=wsi_path,
        coordinates=coordinates,
        tile_level=level,
        tile_size=tiling_params.tile_size,
        resize_factor=resize_factor,
        tile_size_lv0=tile_size_lv0,
        target_spacing=tiling_params.spacing,
        save_dir=coordinates_dir,
    )

    tile_encoder = Virchow(model_dir=model_dir)
    prism = PRISM(model_dir=model_dir)

    caption = generate_caption(
        wsi_path,
        tile_encoder,
        prism,
        coordinates_dir,
        backend="asap",
        batch_size=batch_size,
        num_workers=num_workers,
        mixed_precision=mixed_precision,
    )

    caption = caption[0].replace("</s>", "").strip()
    caption_translated = translate(caption, model_dir, model_name="opus-mt-en-nl")
    save_output(caption_translated, image_name)



from pathlib import Path

import torch
import torch.nn as nn

from unicorn_baseline.vision.pathology.model_utils import update_state_dict
from unicorn_baseline.vision.pathology.titan.configuration_titan import TitanConfig
from unicorn_baseline.vision.pathology.titan.modeling_titan import Titan


class SlideFeatureExtractor(nn.Module):
    def __init__(self, input_size: int = 224):
        self.input_size = input_size
        super(SlideFeatureExtractor, self).__init__()
        self.build_encoders()
        self.set_device()
        for param in self.parameters():
            param.requires_grad = False

    def set_device(self):
        if torch.cuda.is_available():
            self.device = torch.device("cuda")
        else:
            self.device = torch.device("cpu")

    def build_encoders(self):
        raise NotImplementedError

    def get_transforms(self):
        return self.tile_encoder.get_transforms()

    def forward(self, x):
        return self.tile_encoder(x)

    def forward_slide(self, **kwargs):
        return self.slide_encoder(**kwargs)

    def __repr__(self):
        total_params = sum(p.numel() for p in self.parameters())
        trainable_params = sum(p.numel() for p in self.parameters() if p.requires_grad)
        return (
            f"{self.__class__.__name__}\n"
            f"Total Parameters: {total_params / 1e6:.2f}M\n"
            f"Trainable Parameters: {trainable_params / 1e6:.2f}M"
        )


class TITAN(SlideFeatureExtractor):
    def __init__(self, model_dir: Path, input_size: int = 512):
        self.model_dir = model_dir
        super(TITAN, self).__init__(input_size)
        self.features_dim = 768

    def build_encoders(self):

        cfg = TitanConfig()
        self.slide_encoder = Titan(cfg)

        checkpoint_path = self.model_dir / "titan-slide-encoder.pth"
        print(f"Loading slide encoder weights from {checkpoint_path} ...")
        self.slide_encoder_weights = torch.load(checkpoint_path)
        updated_sd, msg = update_state_dict(
            model_dict=self.slide_encoder.state_dict(),
            state_dict=self.slide_encoder_weights
        )
        self.slide_encoder.load_state_dict(updated_sd, strict=True)
        print(msg)

        print(f"Building tile encoder ...")
        self.tile_encoder, self.eval_transform = self.slide_encoder.return_conch(
            self.model_dir
        )

    def get_transforms(self):
        return self.eval_transform

    def forward_slide(self, tile_features, tile_coordinates, tile_size_lv0):
        tile_features = tile_features.unsqueeze(0)
        tile_coordinates = tile_coordinates.unsqueeze(0)
        output = self.slide_encoder.encode_slide_from_patch_features(
            tile_features, tile_coordinates, tile_size_lv0
        )
        return output



import cv2
import numpy as np


class HasEnoughTissue(object):
    def __init__(self, contour, contour_holes, tissue_mask, tile_size, scale, pct=0.01):
        self.cont = contour
        self.holes = contour_holes
        self.mask = tissue_mask // 255
        self.tile_size = tile_size
        self.scale = scale
        self.pct = pct

        # Precompute the combined tissue mask
        self.precomputed_mask = self._precompute_tissue_mask()

    def _precompute_tissue_mask(self):
        """
        Precompute a binary mask for the entire region, combining the contour and holes.

        Returns:
            np.ndarray: A binary mask where tissue regions are 1 and non-tissue regions are 0.
        """
        contour_mask = np.zeros_like(self.mask, dtype=np.uint8)

        # Draw white filled contour on black background
        cv2.drawContours(contour_mask, [self.cont], 0, 255, -1)

        # Draw black filled holes on white filled contour
        cv2.drawContours(contour_mask, self.holes, 0, 0, -1)

        # Combine with the tissue mask
        return cv2.bitwise_and(self.mask, contour_mask)

    def __call__(self, pt):
        """
        Check if a single tile at the given point has enough tissue.

        Args:
            pt (tuple): The (x, y) coordinates of the top-left corner of the tile.

        Returns:
            tuple: (keep_flag, tissue_pct), where:
                - keep_flag is 1 if the tile has enough tissue, otherwise 0.
                - tissue_pct is the percentage of tissue in the tile.
        """
        downsampled_tile_size = int(self.tile_size * 1 / self.scale[0])
        assert (
            downsampled_tile_size > 0
        ), f"downsampled tile_size is equal to zero, aborting; please consider using a smaller seg_params.downsample parameter"
        downsampled_pt = pt * 1 / self.scale[0]
        x_tile, y_tile = map(int, downsampled_pt)

        # Extract the sub-mask for the tile
        sub_mask = self.precomputed_mask[
            y_tile : y_tile + downsampled_tile_size,
            x_tile : x_tile + downsampled_tile_size,
        ]

        tile_area = downsampled_tile_size**2
        tissue_area = np.sum(sub_mask)
        tissue_pct = round(tissue_area / tile_area, 3)

        if tissue_pct >= self.pct:
            return 1, tissue_pct
        else:
            return 0, tissue_pct

    def check_coordinates(self, coords):
        """
        Check multiple tile coordinates for tissue coverage in a vectorized manner.

        Args:
            coords (np.ndarray): An array of shape (N, 2), where each row is (x, y).

        Returns:
            tuple: (keep_flags, tissue_pcts), where:
                - keep_flags is a list of 1s and 0s indicating whether each tile has enough tissue.
                - tissue_pcts is a list of tissue percentages for each tile.
        """
        downsampled_tile_size = int(self.tile_size * 1 / self.scale[0])
        assert (
            downsampled_tile_size > 0
        ), f"downsampled tile_size is equal to zero, aborting; please consider using a smaller seg_params.downsample parameter"

        # Downsample coordinates
        downsampled_coords = coords * 1 / self.scale[0]
        downsampled_coords = downsampled_coords.astype(int)

        keep_flags = []
        tissue_pcts = []

        for x_tile, y_tile in downsampled_coords:
            # Extract the sub-mask for the tile
            sub_mask = self.precomputed_mask[
                y_tile : y_tile + downsampled_tile_size,
                x_tile : x_tile + downsampled_tile_size,
            ]

            tile_area = downsampled_tile_size**2
            tissue_area = np.sum(sub_mask)
            tissue_pct = round(tissue_area / tile_area, 3)

            if tissue_pct >= self.pct:
                keep_flags.append(1)
                tissue_pcts.append(tissue_pct)
            else:
                keep_flags.append(0)
                tissue_pcts.append(tissue_pct)

        return keep_flags, tissue_pcts



import json
import os

import timm
import torch
import torch.nn as nn
from timm.data import resolve_data_config
from timm.data.transforms_factory import create_transform
from timm.layers import SwiGLUPacked

from unicorn_baseline.vision.pathology.model_utils import update_state_dict


class Virchow(nn.Module):
    """
    Tile-level feature extractor.
    """

    def __init__(self, model_dir, input_size=224):
        super().__init__()
        self.model_dir = model_dir
        self.input_size = input_size
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        # Load model configuration
        with open(os.path.join(self.model_dir, "virchow-config.json"), "r") as f:
            self.config = json.load(f)

        if input_size == 256:
            self.config["pretrained_cfg"]["crop_pct"] = (
                224 / 256
            )  # Ensure Resize is 256

        # Initialize tile encoder
        self.tile_encoder = timm.create_model(
            **self.config, mlp_layer=SwiGLUPacked, act_layer=torch.nn.SiLU
        )

        self.load_weights()
        self.transforms = self.get_transforms()

    def load_weights(self):
        """Load pretrained weights for the tile encoder."""
        checkpoint_path = os.path.join(self.model_dir, "virchow-tile-encoder.pth")
        print(f"Loading tile encoder weights from {checkpoint_path}...")
        weights = torch.load(checkpoint_path, map_location=self.device)
        updated_sd, msg = update_state_dict(
            model_dict=self.tile_encoder.state_dict(), state_dict=weights
        )
        print(msg)
        self.tile_encoder.load_state_dict(updated_sd, strict=True)
        self.tile_encoder.to(self.device)
        self.tile_encoder.eval()

    def get_transforms(self):
        """Retrieve the transformation pipeline for input images."""
        data_config = resolve_data_config(
            self.config["pretrained_cfg"], model=self.tile_encoder
        )
        return create_transform(**data_config)

    def forward(self, x):
        """Extract tile-level embeddings."""
        x = x.to(self.device)
        with torch.no_grad():
            output = self.tile_encoder(x)

        # Extract class and patch tokens
        class_token = output[:, 0]
        patch_tokens = output[:, 1:]
        embedding = torch.cat([class_token, patch_tokens.mean(dim=1)], dim=-1)
        return embedding


class PRISM(nn.Module):
    """
    Slide-level feature extractor (PRISM model).
    """

    def __init__(self, model_dir, input_size=224):
        super().__init__()
        self.model_dir = model_dir
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        # Initialize encoder
        self.slide_encoder = self.build_slide_encoder()

    def build_slide_encoder(self):
        """Load slide encoder (PRISM)."""
        import sys

        sys.path.insert(0, self.model_dir)
        from unicorn_baseline.vision_language.prism.configuring_prism import (
            PerceiverConfig,
            PrismConfig,
        )
        from unicorn_baseline.vision_language.prism.modeling_prism import Prism
        from transformers.models.biogpt.configuration_biogpt import BioGptConfig

        cfg = PrismConfig(
            biogpt_config=BioGptConfig(),
            perceiver_config=PerceiverConfig(),
            model_dir=self.model_dir,
        )
        slide_encoder = Prism(cfg)

        checkpoint_path = os.path.join(self.model_dir, "prism-slide-encoder.pth")
        print(f"Loading slide encoder weights from {checkpoint_path}...")
        weights = torch.load(checkpoint_path, map_location=self.device)
        updated_sd, msg = update_state_dict(
            model_dict=slide_encoder.state_dict(), state_dict=weights
        )
        print(msg)
        slide_encoder.load_state_dict(updated_sd, strict=True)
        slide_encoder.to(self.device)
        slide_encoder.eval()

        return slide_encoder

    def forward_slide(self, embeddings):
        """Generate slide-level captions from tile embeddings."""
        embeddings = embeddings.to(self.device)

        reprs = self.slide_encoder.slide_representations(embeddings)

        genned_ids = self.slide_encoder.generate(
            key_value_states=reprs["image_latents"],
            do_sample=False,
            num_beams=5,
            num_beam_groups=1,
        )
        genned_caption = self.slide_encoder.untokenize(genned_ids)

        return genned_caption



#  Copyright 2025 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
import time
from pathlib import Path
from typing import List

import pandas as pd
from dragon_baseline.main import DragonBaseline
from llm_extractinator import extractinate

from unicorn_baseline.io import write_json_file
from unicorn_baseline.language.task_definitions import TASK_DEFINITIONS


def task16_preprocessing(text_parts):
    return "Roman numeral: " + text_parts[0] + "\n\nText:" + text_parts[1]


def setup_folder_structure(
    basepath: Path, test_data: pd.DataFrame, filename: str = "test"
):
    basepath.mkdir(exist_ok=True, parents=True)
    (basepath / "data").mkdir(exist_ok=True)
    (basepath / "output").mkdir(exist_ok=True)
    (basepath / "tasks").mkdir(exist_ok=True)

    test_data.to_json(basepath / "data" / f"{filename}.json", orient="records")


def wait_for_predictions(
    runpath: Path, task_name: str, timeout: int = 300, interval: int = 10
):
    """
    Wait for the predictions to be generated and saved.

    Args:
        timeout (int): Maximum time to wait in seconds.
        interval (int): Interval between checks in seconds.
    """
    start_time = time.time()
    while time.time() - start_time < timeout:
        for folder in runpath.iterdir():
            if task_name in folder.name:
                print(f"Predictions found in {folder}. Proceeding to postprocess.")
                return folder
        print("Waiting for predictions to complete...")
        time.sleep(interval)
    raise TimeoutError(f"Predictions not found within {timeout} seconds.")


def drop_keys_except(data: List, keys: List[str]) -> List:
    """
    Drop all keys from the dictionary except the specified keys.
    """
    return [
        {key: value for key, value in example.items() if key in keys}
        for example in data
    ]


def generate_task_definition(config: DragonBaseline):
    """
    Combines shared config and task-specific information.

    Parameters:
        config (DragonBaseline): Object with attributes: jobid, input_name

    Returns:
        dict: The full task definition
    """
    task_id = f"Task{int(config.jobid):03}"  # e.g., 12 -> "Task012"

    if task_id not in TASK_DEFINITIONS:
        raise ValueError(f"Unknown task ID: {task_id}")

    base = TASK_DEFINITIONS[task_id].copy()
    base.update(
        {
            "Input_Field": config.input_name,
            "Data_Path": "test.json",
        }
    )
    return base


def generate_task_file(config: DragonBaseline, task_folder: Path):
    """
    Generates a task file based on the task ID and configuration.

    Parameters:
        task_id (str): e.g., "17"
        config (dict): dict with keys: 'Input_Field', 'Label_Field', 'Data_Path'
        output_path (Path): Path to save the generated task file
    """
    task_definition = generate_task_definition(config)
    output_path = task_folder / f"Task{int(config.jobid):03}.json"
    with open(output_path, "w") as f:
        json.dump(task_definition, f, indent=4)


def post_process_predictions(data: json, task_config: DragonBaseline):
    task_id = task_config.jobid
    prediction_name = task_config.target.prediction_name
    if prediction_name in (
        "single_label_binary_classification",
        "single_label_regression",
    ):
        for example in data:
            example[prediction_name] = float(example[prediction_name])
    if task_id == 15:
        for example in data:
            example[prediction_name] = [
                example.pop("left"),
                example.pop("right"),
            ]
    elif task_id == 16:
        for example in data:
            keys = [
                "biopsy",
                "cancer",
                "high_grade_dysplasia",
                "hyperplastic_polyps",
                "low_grade_dysplasia",
                "non_informative",
                "serrated_polyps",
            ]
            example[prediction_name] = [
                1.0 if example.pop(key) in ["True", True] else 0.0 for key in keys
            ]
    elif task_id == 17:
        for example in data:
            example[prediction_name] = [
                float(example.pop("lesion_1")),
                float(example.pop("lesion_2")),
                float(example.pop("lesion_3")),
                float(example.pop("lesion_4")),
                float(example.pop("lesion_5")),
            ]
    data = drop_keys_except(data, ["uid", prediction_name])
    return data


def run_language(OUTPUT_PATH: Path) -> int:
    # Read the task configuration, few-shots and test data
    # We'll leverage the DRAGON baseline algorithm for this
    algorithm = DragonBaseline()
    algorithm.load()
    algorithm.analyze()  # needed for verification of predictions
    task_config = algorithm.task
    few_shots = algorithm.df_train
    test_data = algorithm.df_test
    basepath = Path("/opt/app/workdir/language")
    setup_folder_structure(basepath, test_data, filename="test")
    print(f"Task description: {task_config}")

    task_name = task_config.task_name
    task_id = task_config.jobid

    generate_task_file(
        config=task_config,
        task_folder=basepath / "tasks",
    )

    # Task specific preprocessing
    if task_name.startswith("Task16_"):
        test_data["text"] = test_data["text_parts"].apply(task16_preprocessing)

    # Perform data extraction
    extractinate(
        task_id=task_id,
        model_name="phi4",
        num_examples=0,
        temperature=0.0,
        max_context_len="max",
        num_predict=512,
        translate=False,
        data_dir=basepath / "data",
        output_dir=basepath / "output",
        task_dir=basepath / "tasks",
        n_runs=1,
        verbose=False,
        run_name="run",
        reasoning_model=False,
        seed=42,
    )

    # Wait for the predictions to be generated and saved
    runpath = basepath / "output" / "run"
    datafolder = wait_for_predictions(
        runpath=runpath,
        task_name=str(task_id),
        timeout=300,
        interval=10,
    )

    # Load the predictions
    datapath = datafolder / "nlp-predictions-dataset.json"
    with open(datapath, "r") as file:
        predictions = json.load(file)

    predictions = post_process_predictions(predictions, task_config)

    # Save the predictions
    test_predictions_path = OUTPUT_PATH / "nlp-predictions-dataset.json"
    write_json_file(
        location=test_predictions_path,
        content=predictions,
    )

    # Verify the predictions
    algorithm.test_predictions_path = test_predictions_path
    algorithm.verify_predictions()

    print(f"Saved neural representation to {test_predictions_path}")
    return 0



import sys
import warnings
import threading
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import NamedTuple

import cv2
import numpy as np
import tqdm
import wholeslidedata as wsd
from numba import njit
from PIL import Image

from unicorn_baseline.vision.pathology.wsi_utils import HasEnoughTissue

# ignore all warnings from wholeslidedata
warnings.filterwarnings("ignore", module="wholeslidedata")

Image.MAX_IMAGE_PIXELS = 933120000
_printed_warning = False
_warning_lock = threading.Lock()


class FilterParams(NamedTuple):
    """
    Parameters for filtering contours.
    """

    ref_tile_size: int  # reference tile size for filtering
    a_t: int  # contour area threshold for filtering
    a_h: int  # hole area threshold for filtering
    max_n_holes: int  # maximum number of holes allowed


class TilingParams(NamedTuple):
    """
    Parameters for tiling.
    """

    spacing: float  # spacing at which to tile the slide, in microns per pixel
    tolerance: float  # tolerance for matching the spacing (float between 0 and 1, deciding how much spacing can deviate from those specified in the slide metadata)
    tile_size: int  # size of the tiles to extract, in pixels
    overlap: float  # overlap between tiles
    drop_holes: bool  # whether to drop tiles that fall within holes
    min_tissue_ratio: float  # minimum tissue percentage required for a tile
    use_padding: bool  # whether to use padding for tiles at the edges


class WholeSlideImage(object):
    """
    A class for handling Whole Slide Images (wsi) and tile extraction.
    Attributes:
        path (Path): Full path to the wsi.
        name (str): Name of the wsi (stem of the path).
        fmt (str): File format of the wsi.
        wsi (wsd.WholeSlideImage): wsi object.
        spacings (List[float]): List of spacings for each level.
        level_dimensions (List[Tuple[int, int]]): Dimensions at each level.
        level_downsamples (List[Tuple[float, float]]): Downsample factors for each level.
        backend (str): Backend used for opening the wsi (default: "asap").
        mask_path (Path): Path to the segmentation mask.
        mask (wsd.WholeSlideImage): Segmentation mask object.
        seg_level (int): Level for segmentation.
        binary_mask (np.ndarray): Binary segmentation mask as a numpy array.
    """

    def __init__(
        self,
        path: Path,
        mask_path: Path | None = None,
        spacing_at_lvl_0: float | None = None,
        downsample: int = 32,
        backend: str = "asap",
    ):
        """
        Initializes a Whole Slide Image object with optional mask and spacing.

        Args:
            path (Path): Path to the wsi.
            mask_path (Path, optional): Path to the tissue mask, if available. Defaults to None.
            spacing_at_lvl_0 (float, optional): Manually set spacing at level 0, if speficied. Defaults to None.
            downsample (int): Downsample factor for finding best level for tissue segmentation. Defaults to 32.
            backend (str): Backend to use for opening the wsi. Defaults to "asap".
        """

        self.path = path
        self.name = path.stem.replace(" ", "_")
        self.fmt = path.suffix
        self.wsi = wsd.WholeSlideImage(path, backend=backend)

        self._scaled_contours_cache = {}  # add a cache for scaled contours
        self._scaled_holes_cache = {}  # add a cache for scaled holes
        self._level_spacing_cache = {}  # add a cache for level spacings

        self.spacing_at_lvl_0 = spacing_at_lvl_0  # manually set spacing at level 0
        self.spacings = self.get_spacings()
        self.level_dimensions = self.wsi.shapes
        self.level_downsamples = self.get_downsamples()
        self.backend = backend

        self.mask_path = mask_path
        if mask_path is not None:
            self.mask = wsd.WholeSlideImage(mask_path, backend=backend)
            self.seg_level = self.load_segmentation(downsample)
        else:
            self.seg_level = self.segment_tissue(downsample)

    def get_downsamples(self):
        """
        Calculate the downsample factors for each level of the image pyramid.

        This method computes the downsample factors for each level in the image
        pyramid relative to the base level (level 0). The downsample factor for
        each level is represented as a tuple of two values, corresponding to the
        downsampling in the width and height dimensions.

        Returns:
            list of tuple: A list of tuples where each tuple contains two float
            values representing the downsample factors (width_factor, height_factor)
            for each level relative to the base level.
        """
        level_downsamples = []
        dim_0 = self.level_dimensions[0]
        for dim in self.level_dimensions:
            level_downsample = (dim_0[0] / float(dim[0]), dim_0[1] / float(dim[1]))
            level_downsamples.append(level_downsample)
        return level_downsamples

    def get_spacings(self):
        """
        Retrieve the spacings for the whole slide image.

        If the `spacing` attribute is not set, the method returns the original spacings
        from the wsi. Otherwise, it calculates adjusted spacings based on the provided
        `spacing` value and the original spacings.

        Returns:
            list: A list of spacings, either the original or adjusted based on the
            `spacing` attribute.
        """
        if self.spacing_at_lvl_0 is None:
            spacings = self.wsi.spacings
        else:
            spacings = [
                self.spacing_at_lvl_0 * s / self.wsi.spacings[0] for s in self.wsi.spacings
            ]
        return spacings

    def get_level_spacing(self, level: int):
        """
        Retrieve the spacing value for a specified level.

        Args:
            level (int): Level for which to retrieve the spacing.

        Returns:
            float: Spacing value corresponding to the specified level.
        """
        if level not in self._level_spacing_cache:
            self._level_spacing_cache[level] = self.spacings[level]
        return self._level_spacing_cache[level]

    def get_best_level_for_spacing(self, target_spacing: float, tolerance: float, verbose: bool = False):
        """
        Determines the best level in a multi-resolution image pyramid for a given target spacing.

        Ensures that the spacing of the returned level is either within the specified tolerance of the target
        spacing or smaller than the target spacing to avoid upsampling.

        Args:
            target_spacing (float): Desired spacing.
            tolerance (float, optional): Tolerance for matching the target spacing.

        Returns:
            level (int): Index of the best matching level in the image pyramid.
        """
        global _printed_warning

        spacing = self.get_level_spacing(0)
        target_downsample = target_spacing / spacing
        level = self.get_best_level_for_downsample_custom(target_downsample)
        level_spacing = self.get_level_spacing(level)

        # check if the level_spacing is within the tolerance of the target_spacing
        is_within_tolerance = False
        if abs(level_spacing - target_spacing) / target_spacing <= tolerance:
            is_within_tolerance = True
            return level, is_within_tolerance

        # otherwise, look for a spacing smaller than or equal to the target_spacing
        else:
            while level > 0 and level_spacing > target_spacing:
                level -= 1
                level_spacing = self.get_level_spacing(level)
                if abs(level_spacing - target_spacing) / target_spacing <= tolerance:
                    is_within_tolerance = True
                    break

        if not abs(level_spacing - target_spacing) / target_spacing <= tolerance and verbose:
            with _warning_lock:
                if not _printed_warning:
                    print(f"Unable to find a spacing within {tolerance:.0%} of the target spacing ({target_spacing:.2f}). Resampling from {level_spacing:.2f} instead.")
                    _printed_warning = True

        return level, is_within_tolerance

    def get_best_level_for_downsample_custom(self, downsample: float | int):
        """
        Determines the best level for a given downsample factor based on the available
        level downsample values.

        Args:
            downsample (float): Target downsample factor.

        Returns:
            int: Index of the best matching level for the given downsample factor.
        """
        level = int(np.argmin([abs(x - downsample) for x, _ in self.level_downsamples]))
        return level

    def load_segmentation(
        self,
        downsample: int,
        sthresh_up: int = 255,
        tissue_pixel_value: int = 1,
    ):
        """
        Processes the segmentation mask to create a binary mask. The binary mask is resized
        to match the dimensions of the slide at the level corresponding to the desired downsample.

        Args:
            downsample (int): Downsample factor for finding best level for tissue segmentation.
            sthresh_up (int, optional): Upper threshold value for scaling the binary
                mask. Defaults to 255.
            tissue_pixel_value (int, optional): Pixel value in the segmentation mask that
                represents tissue. Defaults to 1.

        Returns:
            int: Level at which the tissue mask was loaded.
        """
        mask_spacing_at_level_0 = self.mask.spacings[0]
        seg_level = self.get_best_level_for_downsample_custom(downsample)
        seg_spacing = self.get_level_spacing(seg_level)

        mask_downsample = seg_spacing / mask_spacing_at_level_0
        mask_level = int(np.argmin([abs(x - mask_downsample) for x in self.mask.downsamplings]))
        mask_spacing = self.mask.spacings[mask_level]

        scale = seg_spacing / mask_spacing
        while scale < 1 and mask_level > 0:
            mask_level -= 1
            mask_spacing = self.mask.spacings[mask_level]
            scale = seg_spacing / mask_spacing

        mask = self.mask.get_slide(spacing=mask_spacing)
        width, height, _ = mask.shape
        # resize the mask to the size of the slide at seg_spacing
        mask = cv2.resize(mask.astype(np.uint8), (int(height / scale), int(width / scale)), interpolation=cv2.INTER_NEAREST)

        m = (mask == tissue_pixel_value).astype("uint8")
        if np.max(m) <= 1:
            m = m * sthresh_up

        self.binary_mask = m
        return seg_level

    def segment_tissue(
        self,
        downsample: int,
        sthresh: int = 20,
        sthresh_up: int = 255,
        mthresh: int = 7,
        close: int = 0,
        use_otsu: bool = False,
    ):
        """
        Segment the tissue via HSV -> Median thresholding -> Binary thresholding -> Morphological closing.

        Args:
            downsample (int): Downsample factor for finding best level for tissue segmentation.
            sthresh (int, optional): Lower threshold for binary thresholding. Defaults to 20.
            sthresh_up (int, optional): Upper threshold for binary thresholding. Defaults to 255.
            mthresh (int, optional): Kernel size for median blurring. Defaults to 7.
            close (int, optional): Size of the kernel for morphological closing.
                If 0, no morphological closing is applied. Defaults to 0.
            use_otsu (bool, optional): Whether to use Otsu's method for thresholding. Defaults to False.

        Returns:
            int: Level at which the tissue mask was created.
        """

        seg_level = self.get_best_level_for_downsample_custom(downsample)
        seg_spacing = self.get_level_spacing(seg_level)

        img = self.wsi.get_slide(spacing=seg_spacing)
        img = np.array(Image.fromarray(img).convert("RGBA"))
        img_hsv = cv2.cvtColor(img, cv2.COLOR_RGB2HSV)  # convert to HSV space
        img_med = cv2.medianBlur(img_hsv[:, :, 1], mthresh)  # apply median blurring

        # thresholding
        if use_otsu:
            _, img_thresh = cv2.threshold(
                img_med, 0, sthresh_up, cv2.THRESH_OTSU + cv2.THRESH_BINARY
            )
        else:
            _, img_thresh = cv2.threshold(
                img_med, sthresh, sthresh_up, cv2.THRESH_BINARY
            )

        # morphological closing
        if close > 0:
            kernel = np.ones((close, close), np.uint8)
            img_thresh = cv2.morphologyEx(img_thresh, cv2.MORPH_CLOSE, kernel)

        self.binary_mask = img_thresh
        return seg_level

    def get_tile_coordinates(
        self,
        tiling_params: TilingParams,
        filter_params: FilterParams,
        num_workers: int = 1,
    ):
        """
        Extract tile coordinates based on the specified target spacing, tile size, overlap,
        and additional tiling and filtering parameters.

        Args:
            tiling_params (NamedTuple, optional): Parameters for tiling, including:
                - "spacing" (float): Spacing at which to tile the slide, in microns per pixel.
                - "tolerance" (float): Tolerance for matching the spacing, deciding how much
                    spacing can deviate from those specified in the slide metadata.
                - "tile_size" (int): Size of the tiles to extract, in pixels.
                - "overlap" (float): Overlap between tiles, as a fraction of the tile size.
                - "drop_holes" (bool): If True, tiles falling within a hole will be excluded. Defaults to False.
                - "min_tissue_ratio" (float, optional): Minimum amount pixels covered with tissue required for a tile. Defaults to 0.25 (25 percent).
                - "use_padding" (bool): Whether to use padding for tiles at the edges. Defaults to True.
            filter_params (NamedTuple, optional): Parameters for filtering contours, including:
                - "ref_tile_size" (int): Reference tile size for filtering.
                - "a_t" (int): Contour area threshold for filtering.
                - "a_h" (int): Hole area threshold for filtering.
                - "max_n_holes" (int): Maximum number of holes allowed.
            num_workers (int, optional): Number of workers to use for parallel processing.
                Defaults to 1.

        Returns:
            Tuple:
                - tile_coordinates (List[Tuple[int, int]]): List of (x, y) coordinates for the extracted tiles.
                - tissue_percentages (List[float]): List of tissue percentages for each tile.
                - tile_level (int): Level of the wsi used for tile extraction.
                - resize_factor (float): The factor by which the tile size was resized.
                - tile_size_lv0 (int): The tile size at level 0 of the wsi pyramid.
        """
        scale = tiling_params.spacing / self.get_level_spacing(0)
        tile_size_lv0 = int(tiling_params.tile_size * scale)

        contours, holes = self.detect_contours(tiling_params.spacing, tiling_params.tolerance, filter_params)
        (
            running_x_coords,
            running_y_coords,
            tissue_percentages,
            tile_level,
            resize_factor,
        ) = self.process_contours(
            contours,
            holes,
            spacing=tiling_params.spacing,
            tolerance=tiling_params.tolerance,
            tile_size=tiling_params.tile_size,
            overlap=tiling_params.overlap,
            drop_holes=tiling_params.drop_holes,
            min_tissue_ratio=tiling_params.min_tissue_ratio,
            use_padding=tiling_params.use_padding,
            num_workers=num_workers,
        )
        tile_coordinates = list(zip(running_x_coords, running_y_coords))
        return (
            tile_coordinates,
            tissue_percentages,
            tile_level,
            resize_factor,
            tile_size_lv0,
        )

    @staticmethod
    def filter_contours(contours, hierarchy, filter_params: FilterParams):
        """
        Filter contours by area using FilterParams.
        """
        filtered = []

        # find indices of foreground contours (parent == -1)
        hierarchy_1 = np.flatnonzero(hierarchy[:, 1] == -1)
        all_holes = []

        # loop through foreground contour indices
        for cont_idx in hierarchy_1:
            # actual contour
            cont = contours[cont_idx]
            # indices of holes contained in this contour (children of parent contour)
            holes = np.flatnonzero(hierarchy[:, 1] == cont_idx)
            # take contour area (includes holes)
            a = cv2.contourArea(cont)
            # calculate the contour area of each hole
            hole_areas = [cv2.contourArea(contours[hole_idx]) for hole_idx in holes]
            # actual area of foreground contour region
            a = a - np.array(hole_areas).sum()
            if a == 0:
                continue
            if a > filter_params.a_t:  # Use named tuple instead of dictionary
                filtered.append(cont_idx)
                all_holes.append(holes)

        foreground_contours = [contours[cont_idx] for cont_idx in filtered]

        hole_contours = []
        for hole_ids in all_holes:
            unfiltered_holes = [contours[idx] for idx in hole_ids]
            unfilered_holes = sorted(
                unfiltered_holes, key=cv2.contourArea, reverse=True
            )
            # take max_n_holes largest holes by area
            unfilered_holes = unfilered_holes[
                : filter_params.max_n_holes
            ]  # Use named tuple
            filtered_holes = []

            # filter these holes
            for hole in unfilered_holes:
                if cv2.contourArea(hole) > filter_params.a_h:  # Use named tuple
                    filtered_holes.append(hole)

            hole_contours.append(filtered_holes)

        return foreground_contours, hole_contours

    def detect_contours(
        self,
        target_spacing: float,
        tolerance: float,
        filter_params: FilterParams,
    ):
        """
        Detect and filter contours from a binary mask based on specified parameters.

        This method identifies contours in a binary mask, filters them based on area
        thresholds, and scales the contours to a specified target resolution.

        Args:
            target_spacing (float): Desired spacing at which tiles should be extracted.
            tolerance (float): Tolerance for matching the target spacing, deciding how much
                spacing can deviate from those specified in the slide metadata.
            filter_params (NamedTuple): A NamedTuple containing filtering parameters:
                - "a_t" (int): Minimum area threshold for foreground contours.
                - "a_h" (int): Minimum area threshold for holes within contours.
                - "max_n_holes" (int): Maximum number of holes to retain per contour.
                - "ref_tile_size" (int): Reference tile size for computing areas.

        Returns:
            Tuple[List[np.ndarray], List[List[np.ndarray]]]:
                - A list of scaled foreground contours.
                - A list of lists containing scaled hole contours for each foreground contour.
        """

        spacing_level, _ = self.get_best_level_for_spacing(target_spacing, tolerance)
        current_scale = self.level_downsamples[spacing_level]
        target_scale = self.level_downsamples[self.seg_level]
        scale = tuple(a / b for a, b in zip(target_scale, current_scale))
        ref_tile_size = filter_params.ref_tile_size  # Use named tuple
        scaled_ref_tile_area = int(ref_tile_size**2 / (scale[0] * scale[1]))

        _filter_params = FilterParams(
            ref_tile_size=filter_params.ref_tile_size,
            a_t=filter_params.a_t * scaled_ref_tile_area,
            a_h=filter_params.a_h * scaled_ref_tile_area,
            max_n_holes=filter_params.max_n_holes,
        )

        # find and filter contours
        contours, hierarchy = cv2.findContours(
            self.binary_mask, cv2.RETR_CCOMP, cv2.CHAIN_APPROX_NONE
        )
        hierarchy = np.squeeze(hierarchy, axis=(0,))[:, 2:]

        foreground_contours, hole_contours = self.filter_contours(
            contours, hierarchy, _filter_params
        )

        # scale detected contours to level 0
        contours = self.scaleContourDim(foreground_contours, target_scale)
        holes = self.scaleHolesDim(hole_contours, target_scale)
        return contours, holes

    @staticmethod
    @njit
    def isInHoles(holes, pt, tile_size: int):
        """
        Check if a given tile is inside any of the specified polygonal holes.

        Args:
            holes (list): A list of polygonal contours, where each contour is represented
                        as a list of points (e.g., from OpenCV's findContours function).
            pt (tuple): The (x, y) coordinates of the top-left corner of the tile to check.
            tile_size (int or float): The size of the tile, used to calculate the center
                                    of the point being tested.

        Returns:
            int: Returns 1 if the point is inside any of the holes, otherwise returns 0.
        """
        for hole in holes:
            if (
                cv2.pointPolygonTest(
                    hole, (pt[0] + tile_size / 2, pt[1] + tile_size / 2), False
                )
                > 0
            ):
                return 1

        return 0

    @staticmethod
    def isInContours(cont_check_fn, pt, drop_holes: bool, tile_size: int, holes: list | None = None):
        """
        Determines whether a given tile is within contours (and optionally outside of holes).

        Args:
            cont_check_fn (callable): A function that checks if a tile is within contours.
                It should accept a (x,y) coordinates as input and return a tuple (keep_flag, tissue_pct),
                where `keep_flag` is a boolean indicating if the tile is within contours,
                and `tissue_pct` is the percentage of tissue coverage of the tile.
            pt (tuple): The (x, y) coordinates of the top-left corner of the tile to check.
                Defaults to None.
            drop_holes (bool): If True, tiles falling within a hole will be excluded.
            tile_size (int): The size of the tile to consider.
            holes (list, optional): A list of holes (e.g., regions to exclude) to check against.

        Returns:
            tuple: A tuple (keep_flag, tissue_pct), where:
                - `keep_flag` is 1 if the tile is within contours and not in holes (if applicable),
                  otherwise 0.
                - `tissue_pct` is the percentage of tissue coverage of the tile.
        """
        keep_flag, tissue_pct = cont_check_fn(pt)
        if keep_flag:
            if holes is not None and drop_holes:
                return not WholeSlideImage.isInHoles(holes, pt, tile_size), tissue_pct
            else:
                return 1, tissue_pct
        return 0, tissue_pct

    @staticmethod
    def scaleContourDim(contours, scale):
        """
        Scales the dimensions of a list of contours by a given factor.

        Args:
            contours (list of numpy.ndarray): A list of contours, where each contour is
                represented as a numpy array of coordinates.
            scale (float): The scaling factor to apply to the contours.

        Returns:
            list of numpy.ndarray: A list of scaled contours, where each contour's
            coordinates are multiplied by the scaling factor and converted to integers.
        """
        return [np.array(cont * scale, dtype="int32") for cont in contours]

    @staticmethod
    def scaleHolesDim(contours, scale):
        """
        Scales the dimensions of holes within a set of contours by a given factor.

        Args:
            contours (list of list of numpy.ndarray): A list of contours, where each contour
                is represented as a list of holes, and each hole is a numpy array of coordinates.
            scale (float): The scaling factor to apply to the dimensions of the holes.

        Returns:
            list of list of numpy.ndarray: A new list of contours with the dimensions of
            the holes scaled by the specified factor.
        """
        return [
            [np.array(hole * scale, dtype="int32") for hole in holes]
            for holes in contours
        ]

    def process_contours(
        self,
        contours,
        holes,
        spacing: float,
        tolerance: float,
        tile_size: int,
        overlap: float,
        drop_holes: bool,
        min_tissue_ratio: float,
        use_padding: bool,
        num_workers: int = 1,
    ):
        """
        Processes a list of contours and their corresponding holes to generate tile coordinates,
        tissue percentages, and other metadata.

        Args:
            contours (list): List of contours representing tissue blobs in the wsi.
            holes (list): List of tissue holes in each contour.
            spacing (float): Desired spacing for tiling.
            tolerance (float): Tolerance for matching the spacing, deciding how much
                spacing can deviate from those specified in the slide metadata.
            tile_size (int): Desired tile size in pixels.
            overlap (float): Overlap between adjacent tiles.
            drop_holes (bool): Whether to drop tiles that fall within holes.
            min_tissue_ratio (float): Minimum amount pixels covered with tissue required for a tile.
            use_padding (bool): Whether to pad the tiles to ensure full coverage.
            num_workers (int, optional): Number of workers to use for parallel processing. Defaults to 1.

        Returns:
            tuple: A tuple containing:
                - running_x_coords (list): The x-coordinates of the extracted tiles.
                - running_y_coords (list): The y-coordinates of the extracted tiles.
                - running_tissue_pct (list): List of tissue percentages for each extracted tile.
                - tile_level (int): Level of the wsi used for tile extraction.
                - resize_factor (float): The factor by which the tile size was resized.
        """
        running_x_coords, running_y_coords = [], []
        running_tissue_pct = []
        tile_level = None
        resize_factor = None

        def process_single_contour(i):
            return self.process_contour(
                contours[i],
                holes[i],
                spacing,
                tolerance,
                tile_size,
                overlap,
                drop_holes,
                min_tissue_ratio,
                use_padding,
            )

        with ThreadPoolExecutor(max_workers=num_workers) as executor:
            results = list(
                tqdm.tqdm(
                    executor.map(process_single_contour, range(len(contours))),
                    desc=f"Extracting tissue tiles",
                    unit=" tissue blob",
                    total=len(contours),
                    leave=True,
                    file=sys.stdout,
                )
            )

        for (
            x_coords,
            y_coords,
            tissue_pct,
            cont_tile_level,
            cont_resize_factor,
        ) in results:
            if len(x_coords) > 0:
                if tile_level is not None:
                    assert (
                        tile_level == cont_tile_level
                    ), "Tile level should be the same for all contours"
                tile_level = cont_tile_level
                resize_factor = cont_resize_factor
                running_x_coords.extend(x_coords)
                running_y_coords.extend(y_coords)
                running_tissue_pct.extend(tissue_pct)

        return (
            running_x_coords,
            running_y_coords,
            running_tissue_pct,
            tile_level,
            resize_factor,
        )

    def process_contour(
        self,
        contour,
        contour_holes,
        spacing: float,
        tolerance: float,
        tile_size: int,
        overlap: float,
        drop_holes: bool,
        min_tissue_ratio: float,
        use_padding: bool = True,
    ):
        """
        Processes a contour to generate tile coordinates and associated metadata.

        Args:
            contour (numpy.ndarray): Contour to process, defined as a set of points.
            contour_holes (list): List of holes within the contour.
            spacing (float): Target spacing for the tiles.
            tolerance (float): Tolerance for matching the spacing, deciding how much
                spacing can deviate from those specified in the slide metadata.
            tile_size (int): Size of the tiles in pixels.
            overlap (float): Overlap between tiles.
            drop_holes (bool): Whether to drop tiles that fall within holes.
            min_tissue_ratio (float): Minimum amount pixels covered with tissue required for a tile.
            use_padding (bool): Whether to pad the image to ensure full coverage.

        Returns:
            tuple: A tuple containing:
                - x_coords (list): List of x-coordinates for each tile.
                - y_coords (list): List of y-coordinates for each tile.
                - filtered_tissue_percentages (list): List of tissue percentages for each tile.
                - tile_level (int): Level of the image used for tile extraction.
                - resize_factor (float): The factor by which the tile size was resized.
        """
        tile_level, is_within_tolerance = self.get_best_level_for_spacing(spacing, tolerance, verbose=True)
        tile_spacing = self.get_level_spacing(tile_level)
        resize_factor = spacing / tile_spacing
        if is_within_tolerance:
            resize_factor = 1.0

        tile_size_resized = int(tile_size * resize_factor)
        step_size = int(tile_size_resized * (1.0 - overlap))

        if contour is not None:
            start_x, start_y, w, h = cv2.boundingRect(contour)
        else:
            start_x, start_y, w, h = (
                0,
                0,
                self.level_dimensions[tile_level][0],
                self.level_dimensions[tile_level][1],
            )

        tile_downsample = (
            int(self.level_downsamples[tile_level][0]),
            int(self.level_downsamples[tile_level][1]),
        )
        ref_tile_size = (
            tile_size_resized * tile_downsample[0],
            tile_size_resized * tile_downsample[1],
        )

        img_w, img_h = self.level_dimensions[0]
        if use_padding:
            stop_y = int(start_y + h)
            stop_x = int(start_x + w)
        else:
            stop_y = min(start_y + h, img_h - ref_tile_size[1] + 1)
            stop_x = min(start_x + w, img_w - ref_tile_size[0] + 1)

        scale = self.level_downsamples[self.seg_level]
        cont = self.scaleContourDim([contour], (1.0 / scale[0], 1.0 / scale[1]))[0]

        tissue_checker = HasEnoughTissue(
            contour=cont,
            contour_holes=contour_holes,
            tissue_mask=self.binary_mask,
            tile_size=ref_tile_size[0],
            scale=scale,
            pct=min_tissue_ratio,
        )

        ref_step_size_x = int(step_size * tile_downsample[0])
        ref_step_size_y = int(step_size * tile_downsample[1])

        x_range = np.arange(start_x, stop_x, step=ref_step_size_x)
        y_range = np.arange(start_y, stop_y, step=ref_step_size_y)
        x_coords, y_coords = np.meshgrid(x_range, y_range, indexing="ij")
        coord_candidates = np.array(
            [x_coords.flatten(), y_coords.flatten()]
        ).transpose()

        # vectorized processing of coordinates using the tissue_checker
        keep_flags, tissue_pcts = tissue_checker.check_coordinates(coord_candidates)

        if drop_holes:
            keep_flags = [
                flag and not self.isInHoles(contour_holes, coord, ref_tile_size[0])
                for flag, coord in zip(keep_flags, coord_candidates)
            ]

        filtered_coordinates = coord_candidates[np.array(keep_flags) == 1]
        filtered_tissue_percentages = np.array(tissue_pcts)[np.array(keep_flags) == 1]

        ntile = len(filtered_coordinates)

        if ntile > 0:
            x_coords = list(filtered_coordinates[:, 0])
            y_coords = list(filtered_coordinates[:, 1])
            return (
                x_coords,
                y_coords,
                filtered_tissue_percentages,
                tile_level,
                resize_factor,
            )

        else:
            return [], [], [], None, None

    @staticmethod
    def process_coord_candidate(
        coord, contour_holes, tile_size, cont_check_fn, drop_holes
    ):
        """
        Processes a candidate coordinate to determine if it should be kept based on
        its location relative to contours and the percentage of tissue it contains.

        Args:
            coord (tuple): (x, y) coordinate to be processed.
            contour_holes (list): A list of contours and holes to check against.
            tile_size (int): Size of the tile to consider.
            cont_check_fn (callable): A function to check if the coordinate is within
                the contours or holes.
            drop_holes (bool): A flag indicating whether to drop tiles falling in holes during the check.

        Returns:
            tuple: A tuple containing:
                - coord (tuple or None): Input coordinate if it passes the check,
                otherwise None.
                - tissue_pct (float): Percentage of tissue in the tile.
        """
        keep_flag, tissue_pct = WholeSlideImage.isInContours(
            cont_check_fn, coord, drop_holes, tile_size, contour_holes
        )
        if keep_flag:
            return coord, tissue_pct
        else:
            return None, tissue_pct

    def get_tile(self, x, y, tile_size, spacing):
        """
        Extracts a tile from a whole slide image at the specified coordinates, size, and spacing.

        Args:
            x (int): The x-coordinate of the top-left corner of the tile.
            y (int): The y-coordinate of the top-left corner of the tile.
            tile_size (tuple): A tuple (width, height) specifying the dimensions of the tile.
            spacing (float): The spacing (resolution) at which the tile should be extracted.

        Returns:
            numpy.ndarray: The extracted tile as a numpy array.
        """
        return self.wsi.get_patch(
            x,
            y,
            tile_size[0],
            tile_size[1],
            spacing=spacing,
            center=False,
        )



#  Copyright 2025 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pathlib import Path
from typing import Any, Iterable

import numpy as np
import SimpleITK as sitk
import torch
import torch.nn.functional as F
from torch import nn
from tqdm import tqdm

from unicorn_baseline.io import resolve_image_path, write_json_file
from unicorn_baseline.vision.radiology.ctfm.ctfm import encode, load_model
from unicorn_baseline.vision.radiology.models import SmallDINOv2
from unicorn_baseline.vision.radiology.patch_extraction import extract_patches
from picai_prep.preprocessing import Sample, PreprocessingSettings

def extract_features_classification(
    model: nn.Module,
    dataset: torch.utils.data.Dataset,
    device: torch.device,
    input_size: int,
    slug: str,
    max_feature_length: int = 4096,
) -> dict:
    model.eval()
    with torch.no_grad():
        batch = dataset[0]
        scan_volume = batch["image"].to(device)

        patch_features = []
        for d in range(scan_volume.shape[0]):
            slice_ = scan_volume[d].unsqueeze(0).unsqueeze(0)
            slice_resized = F.interpolate(
                slice_, size=(input_size, input_size), mode="bilinear", align_corners=False
            )
            slice_3ch = slice_resized.repeat(1, 3, 1, 1).to(device)
            feat = model(slice_3ch)
            patch_features.append(feat.squeeze(0).cpu())

        image_level_feature = torch.stack(patch_features).mean(dim=0)
        feature_list = image_level_feature.tolist()[:max_feature_length]

        return {
            "title": slug,
            "features": feature_list
        }


def extract_features_segmentation(
    image,
    model_dir: str,
    title: str = "patch-level-neural-representation",
    patch_size: list[int] = [16, 224, 224],
    patch_spacing: list[float] | None = None,
) -> list[dict]:
    """
    Generate a list of patch features from a radiology image
    """
    patch_features = []
    
    image_orientation = sitk.DICOMOrientImageFilter_GetOrientationFromDirectionCosines(image.GetDirection())
    if image_orientation != 'SPL': 
        image = sitk.DICOMOrient(image, desiredCoordinateOrientation='SPL')
    

    print(f"Extracting patches from image")
    patches, coordinates = extract_patches(
        image=image,
        patch_size=patch_size,
        spacing=patch_spacing,
    )
    if patch_spacing is None:
        patch_spacing = image.GetSpacing()

    model = load_model(model_dir) 
    print(f"Extracting features from patches")
    for patch, coords in tqdm(zip(patches, coordinates), total=len(patches), desc="Extracting features"):
        patch_array = sitk.GetArrayFromImage(patch)
        features = encode(model, patch_array)
        patch_features.append({
            "coordinates": coords[0],
            "features": features,
        })

    patch_level_neural_representation = make_patch_level_neural_representation(
        patch_features=patch_features,
        patch_size=patch_size,
        patch_spacing=patch_spacing,
        image_size=image.GetSize(),
        image_origin=image.GetOrigin(),
        image_spacing=image.GetSpacing(),
        image_direction=image.GetDirection(),
        title=title,
    )
    return patch_level_neural_representation


def make_patch_level_neural_representation(
    *,
    title: str,
    patch_features: Iterable[dict],
    patch_size: Iterable[int],
    patch_spacing: Iterable[float],
    image_size: Iterable[int],
    image_spacing: Iterable[float],
    image_origin: Iterable[float] = None,
    image_direction: Iterable[float] = None,
) -> dict:
    if image_origin is None:
        image_origin = [0.0] * len(image_size)
    if image_direction is None:
        image_direction = np.identity(len(image_size)).flatten().tolist()
    return {
        "meta": {
            "patch-size": list(patch_size),
            "patch-spacing": list(patch_spacing),
            "image-size": list(image_size),
            "image-origin": list(image_origin),
            "image-spacing": list(image_spacing),
            "image-direction": list(image_direction),
        },
        "patches": list(patch_features),
        "title": title,
    }


def run_radiology_vision_task(
    *,
    task_type: str,
    input_information: dict[str, Any],
    model_dir: Path,
):
    # Identify image inputs
    image_inputs = []
    for input_socket in input_information:
        if input_socket["interface"]["kind"] == "Image":
            image_inputs.append(input_socket)

    if task_type == "classification":
        device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        model = SmallDINOv2(model_dir=model_dir, use_safetensors=True).to(device)

        outputs = []
        for image_input in image_inputs:
            slug = image_input["interface"]["slug"]
            image_dir = Path(image_input["input_location"])
            scan_path = next(image_dir.glob("*.mha"), None)
            if scan_path is None:
                continue

            from unicorn_baseline.vision.radiology.dataset import get_scan_dataset
            dataset = get_scan_dataset(scan_path, seed=42)

            result = extract_features_classification(
                model=model,
                dataset=dataset,
                device=device,
                input_size=518,
                slug=slug
            )
            outputs.append(result)

        output_dir = Path("/output")
        output_path = output_dir / "image-neural-representation.json"
        write_json_file(location=output_path, content=outputs)

    elif task_type in ["detection", "segmentation"]:
        output_dir = Path("/output")
        neural_representations = []

        if image_inputs[0]['interface']['slug'].endswith('prostate-mri'):
            images_to_preprocess = {}
            for image_input in image_inputs:
                image_path = resolve_image_path(location=image_input["input_location"])
                print(f"Reading image from {image_path}")
                image = sitk.ReadImage(str(image_path))

                if 't2' in str(image_input["input_location"]): 
                    images_to_preprocess.update({'t2' : image})
                if 'hbv' in str(image_input["input_location"]): 
                    images_to_preprocess.update({'hbv' : image})
                if 'adc' in str(image_input["input_location"]): 
                    images_to_preprocess.update({'adc' : image})

            pat_case = Sample(scans=[images_to_preprocess.get('t2'), images_to_preprocess.get('hbv'), images_to_preprocess.get('adc')], settings=PreprocessingSettings(spacing=[1,1,1], matrix_size=[16,256,256]))
            pat_case.preprocess()
            
            for image in pat_case.scans:
                neural_representation = extract_features_segmentation(
                        image=image,
                        model_dir=model_dir, 
                        title=image_input["interface"]["slug"]
                )
                neural_representations.append(neural_representation)

        else: 
            for image_input in image_inputs:
                image_path = resolve_image_path(location=image_input["input_location"])
                print(f"Reading image from {image_path}")
                image = sitk.ReadImage(str(image_path))

                neural_representation = extract_features_segmentation(
                    image=image,
                    model_dir=model_dir, 
                    title=image_input["interface"]["slug"]
                )
                neural_representations.append(neural_representation)

        output_path = output_dir / "patch-neural-representation.json"
        write_json_file(location=output_path, content=neural_representations)